/**
 * CMI : Cluster Method Invocation
 * Copyright (C) 2007,2008 Bull S.A.S.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id:CmiConfig.java 914 2007-05-25 16:48:16Z loris $
 * --------------------------------------------------------------------------
 */

package org.ow2.cmi.config;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import net.jcip.annotations.ThreadSafe;

import org.ow2.cmi.controller.server.ServerClusterViewManager;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Utility to create a cluster view manager.
 * @author The new CMI team
 */
@ThreadSafe
public final class CMIConfig {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(CMIConfig.class);

    /**
     * Filename for the properties.
     */
    public static final String CMI_PROPERTIES = "cmi.properties";

    private static final String DEFAULT_SERVER_CONF_FILENAME = "cmi-server-config.xml";


    /*------------ Begin of default values: ----------------*/


    /**
     * JGroupsClusterViewManager is the default manager.
     */
    private static final String DEFAULT_MANAGER = "org.ow2.cmi.controller.server.impl.jgroups.JGroupsClusterViewManager";


    /**
     * Default name for binding an instance of ClientClusterViewProvider.
     */
    private static final String DEFAULT_PROVIDER_BIND_NAME = "client_provider";

    /**
     * Default name for binding an instance of the dummy objects representing
     * registry.
     */
    private static final String DEFAULT_DUMMY_REGISTRY_BIND_NAME = "dummy_registry";

    /**
     * Default values for the load.
     */
    public static final String DEFAULT_DEPLOY_DIRECTORY = "deploy";

    /*------------ End of default values ----------------*/

    /**
     * Configure only once.
     */
    private static boolean configured = false;

    /**
     * True if the CMI Server is automatically started.
     */
    private static boolean autostart = true;

    /**
     * True if the replication is enabled on the server.
     */
    private static boolean replicationEnabled = true;

    /**
     * Manager for the cluster view at the server-side.
     */
    private static String serverClusterViewManagerClassname = DEFAULT_MANAGER;

    /**
     * Name of the ClientClusterViewProvider that is used to bind it.
     */
    private static String bindNameForProvider = DEFAULT_PROVIDER_BIND_NAME;

    /**
     * Path to get the configuration files.
     */
    private static String confDir;

    /**
     * Name of the DummyContext that is used to bind it.
     */
    private static String bindNameForDummyRegistry = DEFAULT_DUMMY_REGISTRY_BIND_NAME;

    /**
     * Associate a protocol with a list of cluster view provider URLs.
     */
    private static Map<String, List<String>> clusterViewProviderUrls;

    /**
     * deploy directory.
     */
    private static String deployDirectory = DEFAULT_DEPLOY_DIRECTORY;

    /**
     * True if CMI is embedded.
     */
    private static boolean embedded = false;

    private static String serverConfFilename = DEFAULT_SERVER_CONF_FILENAME;


    /**
     * Utility class, so not public constructor.
     */
    private CMIConfig() {
    }

    /**
     * Initializes the CMI properties with reading the file cmi.properties.
     *
     * @throws CMIConfigException
     *             if the configuration files are not correct
     */
    public static synchronized void init() throws CMIConfigException {
        setProperties(Thread.currentThread().getContextClassLoader()
                .getResource(CMIConfig.CMI_PROPERTIES));
    }

    /**
     * Initializes the CMI properties with reading the given properties.
     *
     * @param props
     *            properties
     * @throws CMIConfigException
     *             if the configuration files are not correct
     */
    public static synchronized void setProperties(final Properties props)
            throws CMIConfigException {
        // Configure only once
        if (configured) {
            return;
        }
        LOGGER.debug("Starting the configuration of CMI...");
        LOGGER.debug("CMI props: {0}", props.toString());

        String repEnabled = props.getProperty(
                CMIProperty.REPLICATION_ENABLED.getPropertyName());

        if (repEnabled != null) {
            replicationEnabled = Boolean.parseBoolean(repEnabled);
        }

        for (Enumeration<?> e = props.propertyNames(); e.hasMoreElements();) {
            String k = (String) e.nextElement();
            String s = props.getProperty(k);
            if (k.equals(CMIProperty.REPLICATION_MANAGER_CLASS
                    .getPropertyName())) {
                // Set class for manager of cluster views (server-side)
                if (replicationEnabled) {
                    serverClusterViewManagerClassname = s;
                }
            } else if (k.equals(CMIProperty.PROVIDER_BIND_NAME
                    .getPropertyName())) {
                bindNameForProvider = s;
            } else if (k.equals(CMIProperty.CLUSTER_VIEW_PROVIDER_URLS
                    .getPropertyName())) {
                if (!replicationEnabled) {
                    try {
                        clusterViewProviderUrls =
                            ProviderURLsParser.parseMultiprotocol(s);
                    } catch (MalformedURLException e1) {
                        LOGGER.error(
                                "Cannot parse the following urls: {0}.", s, e1);
                        throw new CMIConfigException(
                                "Cannot parse the following urls: " + s, e1);
                    }
                }
            } else if (k.equals(CMIProperty.DEPLOY_DIRECTORY.getPropertyName())) {
                deployDirectory = s;
            } else if (k.equals(CMIProperty.AUTOSTART.getPropertyName())) {
                if (replicationEnabled) {
                    autostart = Boolean.parseBoolean(s);
                }
            } else if(k.equals(CMIProperty.SERVER_CONF_FILENAME.getPropertyName())) {
                if (replicationEnabled) {
                    serverConfFilename = s;
                }
            }
        }
        configured = true;
        LOGGER.debug("Configuration of CMI is done.");
    }

    /**
     * Initializes the CMI properties with reading the file at a given location.
     *
     * @param url
     *            the URL to load CMI configuration
     * @throws CMIConfigException
     *             If the configuration files are not correct.
     */
    public static synchronized void setProperties(final URL url)
            throws CMIConfigException {
        if (url == null) {
            LOGGER.error("The given URL is null");
            throw new CMIConfigException("The given URL is null");
        }
        // Load properties from configuration file
        Properties props = UtilConfig.getPropertiesFromURL(url);
        setProperties(props);
    }

    /**
     * @return The class of the ServerClusterViewManager.
     * @throws ClassNotFoundException
     *             if the class implementing the server-side manager cannot be
     *             loaded
     */
    @SuppressWarnings("unchecked")
    public static Class<? extends ServerClusterViewManager> getServerClusterViewManagerClass()
            throws ClassNotFoundException {
        return (Class<? extends ServerClusterViewManager>) CMIConfig.class
                .getClassLoader().loadClass(serverClusterViewManagerClassname);
    }

    /**
     * @return name of the ClientClusterViewProvider that is used to bind it
     */
    public static String getBindNameForProvider() {
        return bindNameForProvider;
    }

    /**
     * @return the bind name of the dummy objects representing registry.
     */
    public static String getBindNameForDummyRegistry() {
        return bindNameForDummyRegistry;
    }

    /**
     * @return true if CMI is configured
     */
    public static boolean isConfigured() {
        return configured;
    }

    /**
     * @return true if CMI is embedded
     */
    public static boolean isEmbedded() {
        return embedded;
    }

    /**
     * Set that CMI is embedded.
     */
    public static void setEmbedded() {
        embedded = true;
    }

    /**
     * @param protocol
     *            a protocol name
     * @return the list of cluster view provider URLs
     */
    public static List<String> getClusterViewProviderUrls(final String protocol) {
        List<String> res = clusterViewProviderUrls.get(protocol);
        if (res == null) {
            return new ArrayList<String>();
        }
        return res;
    }

    /**
     * @return true if the replication is enabled on the server
     */
    public static boolean isReplicationEnabled() {
        return replicationEnabled;
    }

    /**
     * @return the path to get the configuration files
     */
    public static String getConfDir() {
        return confDir;
    }

    /**
     * Set the path to get the configuration files.
     *
     * @param confDir
     *            the path to get the configuration files
     */
    public static void setConfDir(final String confDir) {
        CMIConfig.confDir = confDir;
    }

    /**
     * @return the deployDirectory
     */
    public static String getDeployDirectory() {
        return deployDirectory;
    }

    /**
     * @return true if the CMI Server is automatically started
     */
    public static boolean isAutoStarted() {
        return autostart;
    }

    public static String getServerConfFilename() {
        return serverConfFilename;
    }

}
